{
This class allows one thread to dump bytes into a buffer where another thread can then pick them up.
The writing thread calls push, which writes a byte into the buffer at the position "writePos".  
This increases the number of "bytesAvailable", so another thread can call "pop" and read the byte from the
position "readPos".  Each read or write increases readPos/writePos until the end of the buffer is reached, then
these values go back to zero.  To prevent bytes from being overwritten before they have been read, a call to "push"
will block until bytesAvailable < buffer-size (calling "pop" to read a byte decreases bytesAvailable)
}
unit cyclicBuffer;

interface

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,inifiles;

type
  TCyclicBuffer = class(TObject)
  private
     size,readPos,writePos,bytesAvailable: integer;
     buffer: array of byte;
     lock: TMultiReadExclusiveWriteSynchronizer;
  protected
  public
    constructor create(_size: integer);
    destructor Destroy; override;
    function pop: byte;
    procedure popBuffer(dest: pbyte; count: integer);
    procedure push(b: byte);
    procedure pushBuffer(src: pbyte; count: integer);
    procedure pushIntegerLittleEndian(number: integer; bytes: integer);
    function popIntegerLittleEndian(bytes: integer): integer;
    function available: integer;
    function getSize: integer;
    procedure unpop(num: integer);
    procedure clear;
  end;


implementation

uses unit1;


constructor TCyclicBuffer.create(_size: integer);
begin
   lock := TMultiReadExclusiveWriteSynchronizer.create;
   size := _size;
   setlength(buffer, size);
   clear;
end;

{destructor}
destructor TCyclicBuffer.Destroy;
begin
   buffer := nil;
   lock.free;
   inherited destroy;
end;


{pop a single byte from the buffer}
function TCyclicBuffer.pop: byte;
begin
   result := 0;
   // wait until a byte is available
   while available = 0 do
   begin
      sleep(100);
      if not connected then exit;
   end;

   lock.BeginWrite;
   if not connected then exit;
   result := buffer[readPos];
   inc(readPos);
   assert(readPos <= size);
   if readPos = size then readPos := 0;
   dec(bytesAvailable);
   assert(bytesAvailable >= 0);
   lock.EndWrite;
end;


{pops a number of bytes into a buffer}
procedure TCyclicBuffer.popBuffer(dest: pbyte; count: integer);
var
   i: integer;
begin
   for i := 1 to count do
   begin
      dest^ := pop;
      inc(dest);
   end;
end;


{push a single byte onto the buffer}
procedure TCyclicBuffer.push(b: byte);
begin
   // wait until there is room for this byte
   while available = size do
   begin
      sleep(100);
      if not connected then exit;
   end;

   lock.BeginWrite;
   if not connected then exit;
   buffer[writePos] := b;
   inc(writePos);
   assert(writePos <= size);
   if writePos = size then writePos := 0;
   inc(bytesAvailable);
   assert(bytesAvailable <= size);
   lock.EndWrite;
end;


{push a number of bytes onto the buffer from a buffer}
procedure TCyclicBuffer.pushBuffer(src: pbyte; count: integer);
var
   i: integer;
begin
   for i := 1 to count do
   begin
      push(src^);
      inc(src);
   end;
end;

{push an integer onto the buffer in little-endian format}
procedure TCyclicBuffer.pushIntegerLittleEndian(number: integer; bytes: integer);
var
   i, multiple: integer;
begin
   multiple := 0;
   for i := 1 to bytes do
   begin
      push((number shr (multiple * 8)) and 255);
      inc(multiple);
   end;
end;

{pop an integer from the buffer in little-endian format}
function TCyclicBuffer.popIntegerLittleEndian(bytes: integer): integer;
var
   i,multiple: integer;
begin
   result :=0;
   multiple := 0;
   for i := 1 to bytes do
   begin
      result := result + ((pop and 255) shl (multiple * 8));
      inc(multiple);
   end;
end;


function TCyclicBuffer.available: integer;
begin
   result := 0;
   lock.BeginRead;
   if not connected then exit;
   assert(bytesAvailable <= size);
   result := bytesAvailable;
   lock.EndRead;
end;


function TCyclicBuffer.getSize: integer;
begin
   result := size;
end;


procedure TCyclicBuffer.unpop(num: integer);
var
   i: integer;
begin
   lock.BeginWrite;
   if not connected then exit;
   for i := 1 to num do
   begin
      dec(readPos);
      if readPos < 0 then readPos := size-1;
   end;
   lock.EndWrite;
end;

procedure TCyclicBuffer.clear;
begin
   bytesAvailable := 0;
   readPos := 0;
   writePos := 0;
end;

end.
